package ssw.mj.test;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import junit.framework.Assert;

import org.junit.Before;

import ssw.mj.Errors;
import ssw.mj.Parser;
import ssw.mj.Scanner;
import ssw.mj.Token;
import ssw.mj.codegen.Decoder;

/**
 * Base class for test cases with utility methods used by all tests.
 */
public abstract class CompilerTestCase {
	public static final String CR = "\r";
	public static final String LF = "\n";

	/** stores the number of the current exercise for simple access by the test cases */
	public static final int EX = Configuration.CURRENT_EXERCISE;
	
	private List<String> expectedErrors;
	private List<String> expectedTokens;
	private List<String> expectedSymTab;
	private byte[] expectedCode;
	private int expectedCodeSize;
	private int expectedDataSize;
	private int expectedMainPC;
	
	private Scanner scanner;
	private Parser parser;
	
	
	@Before
	public void setUp() {
		// initialize expected compiler output
		expectedErrors = new ArrayList<String>();
		expectedTokens = new ArrayList<String>();
		expectedSymTab = new ArrayList<String>();
		expectedCode = new byte[8192];
		expectedCodeSize = 0;
		expectedDataSize = 0;
		expectedMainPC = -1;

		if (Configuration.PRINT_ALL_OUTPUT) {
			// print header for console output
			System.out.println("--------------------------------------------------");
			// TODO: Print the name of the test, the getName functionality is gone with JUnit 4.0
		}
	}


	protected void init(String s) {
		scanner = new Scanner(new StringReader(s));
		parser = new Parser(scanner);
	}


	private List<String> splitString(String s) {
		StringTokenizer st = new StringTokenizer(s, "\n");
		List<String> result = new ArrayList<String>();
		while (st.hasMoreTokens()) {
			result.add(st.nextToken());
		}
		return result;
	}
	
	private void print(String title, List<String> expected, List<String> actual) {
		if (Configuration.PRINT_ALL_OUTPUT || !expected.equals(actual)) {
			System.out.println(title);
			System.out.format("  %-60s %s\n", "expected", "actual");
			int lines = Math.max(expected.size(), actual.size());
			for (int i = 0; i < lines; i++) {
				String expectedLine = (i < expected.size() ? expectedLine = expected.get(i) : "");
				String actualLine =   (i < actual.size()   ? actualLine =   actual.get(i)   : "");
				System.out.format("%s %-60s %s\n", (expectedLine.equals(actualLine) ? " " : "x"), expectedLine, actualLine);
			}
		}
	}

	protected void scanAndVerify() {
		List<String> actualTokens = new ArrayList<String>();
		// scan only the expected number of tokens to prevent endless loops
		for (int i = 0; i < getExpectedTokens().size(); i++) {
			actualTokens.add(scanner.next().toString());
		}

		print("Errors", getExpectedErrors(), getActualErrors());
		print("Tokens", getExpectedTokens(), actualTokens);
		
		Assert.assertEquals("Errors", getExpectedErrors(), getActualErrors());
		Assert.assertEquals("Tokens", getExpectedTokens(), actualTokens);
		Assert.assertTrue("Complete Input Scanned", scanner.next().toString().contains("end of file"));
	}

	protected void parseAndVerify() {
		boolean hasPanicError = false;
		try {
			parser.parse();
		} catch (Errors.PanicMode error) {
			hasPanicError = true;
		}
		
		// print SymTab and Code in Java syntax for easy test case generation
//		System.out.println();
//		for (String s : getActualSymTab()) {
//			System.out.println("\t\texpectSymTab(\"" + s + "\");");
//		}
//		System.out.println();
//		List<String> actual = getActualCode();
//		for (int i = 0; i < actual.size() - 3; i++) {
//			String s = actual.get(i);
//			System.out.println("\t\texpectCode(EX >= 6, Code." + s.substring(s.indexOf(":") + 2, s.length()) + ");");
//		}
//		System.out.println("\t\texpectDataSize(" + Code.dataSize + ");");
//		System.out.println("\t\texpectMainPC(" + Code.mainpc + ");");
//		System.out.println();
		
		// print all output before the first assertion to support debugging
		print("Errors", getExpectedErrors(), getActualErrors());
		if (expectedSymTab.size() > 0) {
			print("Symbol Table", getExpectedSymTab(), getActualSymTab());
		}
		if (expectedCodeSize > 0) {
			print("Code Buffer", getExpectedCode(), getActualCode());
		}
		
		Assert.assertEquals("Errors", getExpectedErrors(), getActualErrors());
		if (expectedSymTab.size() > 0) {
			Assert.assertEquals("Symbol Table", getExpectedSymTab(), getActualSymTab());
		}
		if (expectedCodeSize > 0) {
			Assert.assertEquals("Code Buffer", getExpectedCode(), getActualCode());
		}
		if (EX <= 3) {
			Assert.assertEquals("panic mode error", getExpectedErrors().size() > 0, hasPanicError);
		} else {
			// with panic mode, the complete input is not scanned, so cannot check this if EX <= 3
			Assert.assertTrue("Complete Input Scanned", scanner.next().toString().contains("end of file"));
		}
	}


	private List<String> getExpectedErrors() {
		return expectedErrors;
	}
	
	private List<String> getActualErrors() {
		return splitString(scanner.errors.dump());
	}
	
	private List<String> getExpectedTokens() {
		return expectedTokens;
	}
	
	private List<String> getExpectedSymTab() {
		return expectedSymTab;
	}
	
	private List<String> getActualSymTab() {
		return splitString(parser.tab.dump());
	}
	
	private List<String> getExpectedCode() {
		Decoder dec = new Decoder();
		List<String> result = splitString(dec.decode(expectedCode, 0, expectedCodeSize));
		result.add("#CodeSize: " + expectedCodeSize);
		result.add("#DataSize: " + expectedDataSize);
		result.add("#MainPC: " + expectedMainPC);
		return result;
	}

	private List<String> getActualCode() {
		return splitString(parser.code.dump());
	}
	
	
	protected void expectError(boolean enabled, int line, int col, Errors.Message msg, Object... msgParams) {
		if (enabled) {
			expectedErrors.add("-- line " + line + " col " + col + ": " + msg.format(msgParams));
		}
	}

	
	protected void expectToken(Token.Kind kind, int line, int col) {
		expectedTokens.add("line " + line + ", col " + col + ", kind " + kind);
	}
	protected void expectToken(Token.Kind kind, int line, int col, String str) {
		expectedTokens.add("line " + line + ", col " + col + ", kind " + kind + ", str " + str);
	}
	protected void expectToken(Token.Kind kind, int line, int col, int val) {
		expectedTokens.add("line " + line + ", col " + col + ", kind " + kind + ", val " + val);
	}
	protected void expectToken(Token.Kind kind, int line, int col, char val) {
		expectedTokens.add("line " + line + ", col " + col + ", kind " + kind + ", val '" + val + "'");
	}
	
	
	protected void expectSymTab(String line) {
		if (EX >= 4) {
			expectedSymTab.add(line);
		}
	}

	protected void expectSymTabUniverse() {
		// first part of the symbol table (universe) that is equal for all programs
		expectSymTab("-- begin scope (0 variables) --");
		expectSymTab("Type int: int");
		expectSymTab("Type char: char");
		expectSymTab("Constant: class(0) null = 0");
		expectSymTab("Method: char chr(1) at 0");
		expectSymTab("  Local Variable 0: int i");
		expectSymTab("Method: int ord(1) at 0");
		expectSymTab("  Local Variable 0: char ch");
		expectSymTab("Method: int len(1) at 0");
		expectSymTab("  Local Variable 0: void[] arr");
	}


	private void expectCode1(int x) {
		expectedCode[expectedCodeSize++] = (byte) x;
	}
	private void expectCode2(int x) {
		expectCode1(x >> 8); 
		expectCode1(x); 
	}

	private void expectCode4(int x) {
		expectCode2(x >> 16); 
		expectCode2(x); 
	}
	
	protected void expectCode(boolean enabled, int op, Object... params) {
		if (enabled) {
			expectCode1(op);
			for (Object param : params) {
				if (param instanceof Byte) {
					expectCode1((Byte)param);
				} else if (param instanceof Short) {
					expectCode2((Short)param);
				} else if (param instanceof Integer) {
					expectCode4((Integer)param);
				} else {
					throw new Error("invalid parameter");
				}
			}
		}
	}

	protected void expectDataSize(int size) {
		if (EX >= 6) {
			expectedDataSize = size;
		}
	}

	protected void expectMainPC(int pc) {
		if (EX >= 6) {
			expectedMainPC = pc;
		}
	}
}
